const semver = require('semver');

module.exports = context => {
  return Promise
    .resolve(context.names)
    .then(names => ({
      // Package JSON
      json: context.json,
      // List of package names having @wix prefix [ '@wix/package' ]
      names,
      // Mapping of migrated packages { 'package': '@wix/package' }
      changes: {},
      // List of outdated packages, which could not be migrated, cause old version was used { 'package': '2.0.0.' }
      outdated: {}
    }))
    .then(payload => {
      if (payload.json.migrateToScopedPackages === false) {
        return payload;
      }
      return Promise
        .resolve(payload)
        .then(payload => updatePackageName(payload))
        .then(payload => updateNonPublishablePackageName(payload))
        .then(payload => updateDependencies(payload, 'dependencies', context.checkVersions, context.versions))
        .then(payload => updateDependencies(payload, 'devDependencies', context.checkVersions, context.versions))
        .then(payload => updateDependencies(payload, 'peerDependencies', context.checkVersions, context.versions))
        .then(payload => updateDependencies(payload, 'optionalDependencies', context.checkVersions, context.versions))
        .then(payload => updateBundledDependencies(payload))
        .then(payload => updateYoshiClientProjectName(payload))
        .then(payload => updateYoshiExternals(payload));
    });
};

function updateNonPublishablePackageName(payload) {
  const json = payload.json;
  const changes = payload.changes;
  const update = [
    (json.private === true || typeof json.private === 'undefined'),
    !json.name.startsWith('@'),
    !((json.publishConfig || {}).registry || '')
  ].every(truthy => truthy);
  if (update) {
    changes[json.name] = `@wix/${json.name}`;
    json.name = changes[json.name];
  }
  return payload;
}

function updatePackageName(payload) {
  const json = payload.json;
  const changes = payload.changes;
  const inWixRegistry = ((json.publishConfig || {}).registry || '').includes('.wix');

  if (!json.private && !json.name.startsWith('@') && inWixRegistry) {
    changes[json.name] = `@wix/${json.name}`;
    json.name = changes[json.name];
    json.publishConfig['@wix:registry'] = 'https://registry.npmjs.org/';
  }
  return payload;
}

function getDependenciesToUpdate(payload, type, checkVersions, versions) {
  const json = payload.json;
  const names = payload.names;
  const outdated = payload.outdated;

  const deps = Object
    .keys(json[type] || {})
    .filter(name => !name.startsWith('@'))
    .filter(name => names.indexOf(`@wix/${name}`) !== -1);

  if (!checkVersions) {
    return Promise.resolve(deps);
  }

  return deps.reduce((acc, curr) => {
    return acc.then(_ => {
      const version = json[type][curr];
      if (version === '*' || version === 'latest') {
        return _.concat(curr);
      }
      return getDependencyVersions(curr, versions)
        .then(versions => {
          const valid = versions.some(candidate => semver.satisfies(candidate, version));
          if (valid) {
            return _.concat(curr);
          }
          outdated[curr] = version;
          return _;
        })
        .catch(() => _);
    });
  }, Promise.resolve([]));
}

function getDependencyVersions(name, versions) {
  return Promise.resolve(versions[`@wix/${name}`] || []);
}

function updateDependencies(payload, type, checkVersions, versions) {
  const json = payload.json;
  const changes = payload.changes;

  return getDependenciesToUpdate(payload, type, checkVersions, versions)
    .then(deps => {
      deps.forEach(name => {
        changes[name] = `@wix/${name}`;
        json[type][`@wix/${name}`] = json[type][name];
        delete json[type][name];
      });
      return payload;
    })
    .catch(() => payload);
}

function updateBundledDependencies(payload) {
  const json = payload.json;
  const changes = payload.changes;
  const names = payload.names;
  if (json.bundledDependencies) {
    json.bundledDependencies = json.bundledDependencies.map(name => {
      if (names.indexOf(`@wix/${name}`) !== -1) {
        changes[name] = `@wix/${name}`;
        return `@wix/${name}`;
      }
      return name;
    });
  }
  return payload;
}

function updateYoshiClientProjectName(payload) {
  const json = payload.json;
  const changes = payload.changes;
  const names = payload.names;
  const clientProjectName = (json.yoshi && json.yoshi.clientProjectName) || '';
  if (clientProjectName) {
    if (names.indexOf(`@wix/${clientProjectName}`) !== -1) {
      json.yoshi.clientProjectName = `@wix/${clientProjectName}`;
      changes[clientProjectName] = json.yoshi.clientProjectName;
    }
  }
  return payload;
}

function updateYoshiExternals(payload) {
  const json = payload.json;
  if (json.yoshi && json.yoshi.externals) {
    const externals = json.yoshi.externals;
    Object.keys(externals).forEach(key => {
      if (payload.changes[key]) {
        externals[`@wix/${key}`] = externals[key];
        delete externals[key];
      }
    });
  }
  return payload;
}
